//************************************************************
// NUOVO CORSO PROGRAMMAZIONE MICROCONTROLLORI PIC IN C
// (C) Bernardo Giovanni
// https://www.settorezero.com
//
// Tutti i diritti riservati, vietata la
// ridistribuzione e/o riutilizzo anche parziale
// del codice senza consenso scritto dell'autore.
// Al seguente codice si applicano tutte le regole gi
// esposte nell'informativa all'indirizzo:
// https://www.settorezero.com/wordpress/info/termini-di-utilizzo-e-privacy/
//
//************************************************************
// ESEMPIO 6A
// Utilizzo modulo ADCC
// realizzazione di un termometro con display a 7 segmenti
//************************************************************
//
// modulo:      main.c
// autore:      Bernardo Giovanni
// mcu:         PIC16F18887
// clock:      	oscillatore interno 32MHz
// compiler:    Microchip XC8 (1.45)
// ide:         MPLAB X (4.15)
// programmer:  PICkit 4
//
//************************************************************
#include "conf_16F18877.h"
#define _XTAL_FREQ 32000000

#include <xc.h>
#include <stdint.h>
#include <stdbool.h>
#include "settings.h"

// Variabili globali
int8_t muxcount=0; // contatore multiplex per far rimanere impresse le cifre
int8_t actdisplay=0; // numero del singolo display correntemente visualizzato (0,1,2,3) 0=display unit (quello pi a destra)
uint16_t temperature=0; // valore di temperatura moltiplicato 10
uint8_t sign=false; // segno temperatura (1=temperatura negativa)

// porte I/O
#define DISPLAY_REG     PORTD 		// Banco al quale sono collegati i singoli led dei display
#define DISPLAY_M       PORTAbits.RA5 // anodo/catodo comune display unit
#define DISPLAY_C       PORTEbits.RE0 // anodo/catodo comune display decine
#define DISPLAY_D       PORTEbits.RE1 // anodo/catodo comune display centinaia
#define DISPLAY_U       PORTEbits.RE2 // anodo/catodo comune display migliaia

// con un display ad anodo comune, i segmenti si accenderanno
// mandando un livello logico 0 ai singoli catodi e sulla base
// del transistor PNP che pilota l'anodo comune.
#define TURN_ON 0
#define TURN_OFF 1

/*
 * definizione simboli per segmenti
 * PIN		SEGMENTO
 * RD0		A
 * RD1		B
 * RD2		C
 * RD3		D
 * RD4		E
 * RD5		F
 * RD6		G
 * RD7      Decimal Point
 * 
 *    A
 * F|---|
 *  | G |B
 *   ---
 * E|   |C
 *  |---|
 *    D
 * 
 * Ad ogni segmento (Sn) associo il valore numerico del bit
 * della porta alla quale il segmento stesso  collegato
 * (es.: Il segmento A  collegato su RD0, quindi sul bit 0 della porta
 *  che vale 1, il segmento E  collegato su RD4 => bit 4 => valore 16 ecc
 */
#define	SA	1
#define	SB	2
#define	SC	4
#define	SD	8
#define	SE	16
#define	SF	32
#define	SG	64
// qui nel mio esempio ho collegato a RD7 all'anodo comune del puntino decimale
// presente in basso a destra per ogni cifra
#define DOT 128 

// Segmenti da accendere in base al numero da visualizzare
// esempio: il numero 1  fatto dai segmenti B e C
#define N0 SA+SB+SC+SD+SE+SF
#define N1 SB+SC
#define N2 SA+SB+SG+SE+SD
#define N3 SA+SB+SG+SC+SD
#define N4 SF+SG+SB+SC
#define N5 SA+SF+SG+SC+SD
#define N6 SA+SF+SG+SE+SC+SD
#define N7 SA+SB+SC
#define N8 SA+SB+SF+SG+SC+SD+SE
#define N9 SA+SB+SF+SG+SC+SD
#define NM SG                   // segno meno

/*
 * Array con i valori da assegnare al registro PORTD per far accendere i segmenti giusti
 * in base al numero da visualizzare.
 * I segmenti sono 7 e tutti collegati ad una stessa porta (PORT D)
 * Definisco i valori in un array in maniera tale da facilitarmi il compito: mi baster
 * semplicemente assegnare alla porta l'elemento N-esimo dell'array quando vorr far apparire il numero N.
 * Es.: voglio far apparire il numero 7 => PORTD=dispnum[N7]
 * L'array lo definisco come costante (const), dal momento che non varia nell'esecuzione
 * del programma, inoltre in questo modo viene memorizzato nella memoria programma e non nella ram
*/
const uint8_t dispnum[]={N0,N1,N2,N3,N4,N5,N6,N7,N8,N9};

// prototipi funzione
void readAnalog(uint8_t channel);
void __interrupt isr(void);

int main(void)
    {
    SYSTEM_init();
    // abilito gli interrupts
    // il bit GIE  definito uguale per tutti i PICmicro.
    // Il Timer0 non  considerato una periferica (vedi schema interrupt sui datasheet)
    // per cui non serve a nulla abilitare il bit PIE
    GIE=1; 
    
    while(1)
        {
        readAnalog(0); // eseguo la lettura sul canale AN0
        __delay_ms(250);
        __delay_ms(250);
        }// while(1)
    }// main

// interrupt vector
void __interrupt isr(void)
    {
    if (PIR0bits.TMR0IF) // L'interrupt  stato causato da un overflow del timer0 ?
        {
        // Ricarico subito il Timer0
        TMR0L=8; //6+2
        uint16_t result=0; // variabile che conterr la singola cifra da visualizzare
        /*
         * Multiplexing display
         * Visualizzo sui display il valore attuale della variabile "temperature".
         * Per fare questo devo accendere alternativamente prima un display
         * e la volta successiva l'altro. L'accensione di uno o dell'altro
         * display  comandata dalla variabile "actdisplay" la quale viene incrementata
         * ogni "muxcount" interrupt. In base al display da visualizzare al momento 
		 * (unit, decine, centinaia, migliaia) devo estrarre dalla variabile soltanto la 
		 * cifra che mi interessa (solo il decimale, solo l'unit ecc)
         * e questo lo faccio usando divisione e operatore mod (%) che mi restituisce il resto della
		 * divisione. Il valore da mostrare sul display correntemente attivo viene 
		 * immagazzinato nella variabile "result"
         */
        switch (actdisplay) // quale dei 4 display  attivo?
            {
            case 0: // display unit, qui visualizzo il decimale della temperatura
                DISPLAY_D=TURN_OFF;
                DISPLAY_C=TURN_OFF;
                DISPLAY_M=TURN_OFF;
                DISPLAY_U=TURN_ON;
                // es: temperature=234
				// 234%10 => 4 (cifra da visualizzare)
                result=temperature%10;
                break;

            case 1: // display decine, qui visualizzo l'unit della temperatura + il puntino decimale che gestisco dopo
                DISPLAY_U=TURN_OFF;
                DISPLAY_C=TURN_OFF;
                DISPLAY_M=TURN_OFF;
                DISPLAY_D=TURN_ON;
				// es: temperature=234
				// 234/10 => 23 (la divisione di interi in C non fornisce decimali!)
				// 23%10 => 3 (cifra da visualizzare)
                result=(temperature/10U)%10; 
            break;

            case 2: // display centinaia, qui visualizzo la decina della temperatura
                DISPLAY_U=TURN_OFF;
                DISPLAY_D=TURN_OFF;
                DISPLAY_M=TURN_OFF;
                result=(temperature/100U)%10;
                // accendo la cifra solo se c' da mostrare le centinaia
                if (temperature>99U)
                    {
                    DISPLAY_C=TURN_ON;
                    }
            break;

            case 3: // display migliaia, qui visualizzo il segno - in caso di temperatura negativa
                // oppure le centinaia della temperatura
                DISPLAY_U=TURN_OFF;
                DISPLAY_D=TURN_OFF;
                DISPLAY_C=TURN_OFF;
                // accendo il digit solo se c' segno o c' temperatura superiore a 99.9 gradi
                // C' il segno
                if (sign)
                    {
                    DISPLAY_M=TURN_ON;
                    }
                else // non c' il segno
                    {
                    // ci sono le centinaia
                    if (temperature>999U)
                        {
                        DISPLAY_M=TURN_ON;
                        result=(temperature/1000U)%10;
                        }
                    }
            break;
            }// switch (actdisplay)

	// Ora imposto il registro sul quale sono collegati i display.
	// I segmenti da accendere in base al numero da visualizzare sono
	// contenuti nell'array "dispnum".

    // display da accendere: quello delle migliaia. C' il segno?
    if (sign && actdisplay==3)
        {
        // segno '-' sul display delle migliaia. Nego perch i segmenti si attivano con livello logico 0 (anodo comune)
        DISPLAY_REG=~(NM);
        }
    else // tutti gli altri casi, compreso display delle migliaia ma senza il segno
        {
        uint8_t num=dispnum[(unsigned)result];
        // se il display  quello delle decine, attivo anche il puntino decimale
        // (rappresentazione a virgola fissa)
        if (actdisplay==1) num+=DOT;
        DISPLAY_REG=(unsigned)~(num); // per l'anodo comune nego il valore
        }
	// incremento il contatore per cambiare il display da accendere.
    // in base al valore raggiunto da muxcount, cambio display
	// ogni display rimane acceso il tempo di 4 interrupt (4mS)
    muxcount++; 
	if (muxcount==4)
            {
            muxcount=0; // azzero il contatore
            actdisplay++;// cambio il display per il prossimo ciclo
            if (actdisplay>3)
                {
				// torno al primo display
                actdisplay=0;
                }
			// attivo il display mandando il segnale al transistor che pilota il comune
            DISPLAY_REG=0xFF; // azzero il display attualmente visualizzato per prevenire l'effetto "ghosting"
            }
        // Resetto il flag interrupt su timer 0
        PIR0bits.TMR0IF=0;
        } // fine che interrupt verificatosi su timer0
    } // fine interrupt

void readAnalog(uint8_t channel)
    {
    int32_t temp=0; // valore temporaneo di temperatura
    //int32_t ADaccumulator=0; // valore nel registro ADACC
    uint8_t i; // counter
    ADPCH=channel;
    // eseguo 16 burst per avere una lettura a display pi stabile
    for (i=0; i<16; i++)
        {
        // avvio la conversione. Da impostazioni del modulo ADCC
        // ogni trigger eseguir 32 misurazioni
        ADCON0bits.ADGO=1;
        // apetto che termina il burst
        while (ADCON0bits.ADGO) {continue;}
		// ADaccumulator=(ADACCH<<8)+ADACCL; // valore del registro ADACC
        // temp += (ADaccumulator>>5); // media fatta a mano
        // in ADFLTR  contenuta la media delle 32 acquisizioni in modalit burst
		temp += ((int32_t)(ADFLTRH<<8)+ADFLTRL);
        // temp+=(uint16_t)((ADRESH<<8)+ADRESL); // lettura da usare solo in modalit legacy
        __delay_ms(10);
        } // ciclo ripetuto 16 volte
	temp >>= 4; // shift destro di 4 posizioni = divisione per 16

	// il modulo ADCC  a 10 bit e viene utilizzato come riferimento l'FVR impostato a 1024mV
	// questo vuol dire che ogni bit vale 1mV (1024mV/2^10)=1mV/Bit
	// questo sensore di temperatura (l'MCP9700) ha questi parametri:
	// range di temperatura: -10 / +125C
	// risoluzione: 10mV/C
	// tensione restituita a 0C = 500mV
	// qui devo convertire il valore in bit restituito dal modulo AD in valore di temperatura espresso in C
	// dalla lettura sottraggo 500mV per riportarmi a zero
    sign=false;
    temp -= 500U;
    if (temp<0)
        {
        sign=true;
        temp *= (-1);
        }

    // dovrei dividere per 10mV/C / 1mV/Bit = 10bit/C
    // ma non lo faccio per non perdere il decimale
    temperature=(unsigned)temp;
    }